class ProviderError(Exception):
    """
    Exceptions that occur whilst contacting an external storage provider's
    API.
    """


class ProviderAuthenticationError(ProviderError):
    """
    Unable to authenticate.
    """


class ProviderGetUploadUrlError(ProviderError):
    """
    Unable to get URL for uploading a file to.
    """


class ProviderAuthorizationError(ProviderError):
    """
    Not authorised to perform the action.
    """


class ProviderFileUploadError(ProviderError):
    """
    Unable to upload file.
    """



"""
Interface for communicating asynchronously with the Backblaze B2 Cloud
Storage API.
"""
import hashlib
import os
import urllib.parse

import aiohttp

from .exceptions import ProviderAuthorizationError, ProviderGetUploadUrlError


class Backblaze:
    API_NAME = 'b2api/'
    API_VERSION = 'v1/'
    API_DOMAIN = 'https://api.backblazeb2.com'
    API_ENDPOINTS = {
        'list_buckets': 'b2_list_buckets/',
        'get_upload_url': 'b2_get_upload_url/',
        'authorize_account': 'b2_authorize_account/',
    }

    def __init__(self, credentials):
        self.account_id = credentials['account_id']
        self.app_key = credentials['app_key']
        self.authorized_base_url = None
        self.authorization_token = None
        self.authorized_session = None

    def _get_api_url(self, action):
        """
        Generate API endpoint URL.

        :param str action: API action to get URL for.
        :return: API endpoint URL.
        :rtype: str
        """
        path = f'{self.API_NAME}{self.API_VERSION}{self.API_ENDPOINTS[action]}'
        if self.authorized_base_url is None:
            return urllib.parse.urljoin(self.API_DOMAIN, path)
        else:
            return urllib.parse.urljoin(self.authorized_base_url, path)

    async def authenticate(self):
        """
        Authenticate to the API and update authorization attributes.

        :raise ClientResponseError: If HTTP status code >= 400.
        :return: JSON API response containing authorization details.
        :rtype: dict
        """
        url = self._get_api_url('authorize_account')
        auth = aiohttp.BasicAuth(self.account_id, self.app_key)
        required = ('apiUrl', 'authorizationToken')
        async with aiohttp.ClientSession(auth=auth) as session:
            async with session.get(url, timeout=30) as response:
                response.raise_for_status()
                response_js = await response.json()
                if all(r in response_js for r in required):
                    self.authorized_base_url = response_js['apiUrl']
                    self.authorization_token = response_js['authorizationToken']  # noqa
                    self.authorized_session = aiohttp.ClientSession(
                        headers={'Authorization': self.authorization_token})
                    return response_js

    async def _get_upload_url(self, bucket_id):
        """
        Retrieve URL used for uploading a file.

        :param bucket_id: bucket to upload file to.
        :raise ClientResponseError: If HTTP status code >= 400.
        :return: JSON API response containing upload URL.
        :rtype: dict
        """
        if self.authorized_session is None:
            raise ProviderAuthorizationError
        url = self._get_api_url('get_upload_url')
        required = ('uploadUrl', 'authorizationToken')
        async with self.authorized_session as session:
            async with session.post(
                    url, json={'bucketId': bucket_id}) as response:
                response.raise_for_status()
                response_js = await response.json()
                if all(r in response_js for r in required):
                    return response_js

    async def upload_file(self, bucket_id, file_to_upload, content_type):
        """
        Upload file.

        :param bucket_id: bucket to upload file to.
        :param file_to_upload: path of file to upload.
        :param content_type: content (MIME) type of file to upload.
        :return: JSON API response containing confirmation of file upload.
        :rtype: dict
        """
        upload_info = await self._get_upload_url(bucket_id)
        if not upload_info:
            raise ProviderGetUploadUrlError
        upload_url = upload_info['uploadUrl']
        upload_token = upload_info['authorizationToken']
        with open(file_to_upload, 'rb') as f:
            file_data = f.read()
        upload_headers = {
            'Authorization': upload_token,
            'X-Bz-File-Name': urllib.parse.quote(
                os.path.basename(file_to_upload)),
            'Content-Type': content_type,
            'X-Bz-Content-Sha1': hashlib.sha1(file_data).hexdigest()
        }
        async with aiohttp.ClientSession(
                headers=upload_headers) as session:
            async with session.post(
                    upload_url, data=file_data) as response:
                response.raise_for_status()
                response_js = await response.json()
                return response_js



from .providers import PROVIDERS


class BlobStorageError(Exception):
    """
    Base exception class for `BlobStorage`.
    """


class BlobStorageUnrecognizedProviderError(BlobStorageError):
    """
    Unrecognized object storage provider.
    """
    def __str__(self):
        return ('Unrecognized object storage provider. Please select one of'
                f' {", ".join(PROVIDERS)}')



from setuptools import find_packages, setup


package_name = 'aiostorage'
long_description = (
    'Interface for performing common object storage operations '
    'asynchronously. The aim is to support multiple object storage '
    'providers, e.g. Google Cloud, Backblaze, etc.'
)
version = '0.1.5'
classifiers = [
    'Development Status :: 1 - Planning',

    'Intended Audience :: Developers',

    'Programming Language :: Python :: 3.6',
]
requirements = (
    'aiohttp>=2.0,<3.0',  # aiohttp>=3.0 requires python>=3.5.3 but readthedocs
    # runs on python3.5
)
project_urls = {
    'Wiki': 'https://family-guy.github.io/aiostorage-wiki/',
    'Source': 'https://github.com/family-guy/aiostorage.git',
    'Tracker': 'https://github.com/family-guy/aiostorage/issues',
    'Documentation': 'http://aiostorage.readthedocs.io/',
}

setup(
    name=package_name,
    description='Asynchronous object storage',
    long_description=long_description,

    version=version,
    packages=find_packages(exclude=('tests', )),

    install_requires=requirements,
    classifiers=classifiers,

    author='Guy King',
    author_email='grking8@gmail.com',
    license='MIT',
    url='https://github.com/family-guy/aiostorage.git',
    project_urls=project_urls,
)



import asyncio

from .exceptions import BlobStorageUnrecognizedProviderError
from .providers import (Backblaze, ProviderAuthenticationError,
                        ProviderFileUploadError, PROVIDERS, )


class BlobStorage:
    PROVIDER_ADAPTER = {
        'backblaze': {
            'adapter': Backblaze,
            'required': ('account_id', 'app_key'),
        }
    }

    def __init__(self, provider, credentials):
        if provider not in PROVIDERS:
            raise BlobStorageUnrecognizedProviderError
        if not all(r in credentials.keys()
                   for r in self.PROVIDER_ADAPTER[provider]['required']):
            raise KeyError
        self.provider = self.PROVIDER_ADAPTER[provider]['adapter'](credentials)
        self.loop = asyncio.get_event_loop()

    async def _upload_file(self, bucket_id, file_to_upload):
        auth_response = await self.provider.authenticate()
        if not auth_response:
            raise ProviderAuthenticationError
        upload_file_response = await self.provider.upload_file(
            bucket_id, file_to_upload['path'], file_to_upload['content_type'])
        if not upload_file_response:
            raise ProviderFileUploadError
        return upload_file_response

    def upload_files(self, bucket_id, files_to_upload):
        async def _upload_files():
            futures = []
            for file_to_upload in files_to_upload:
                future = asyncio.ensure_future(
                    self._upload_file(bucket_id, file_to_upload))
                futures.append(future)
            return await asyncio.gather(*futures)
        return self.loop.run_until_complete(_upload_files())



from .backblaze import Backblaze
from .exceptions import (ProviderAuthenticationError,
                         ProviderAuthorizationError,
                         ProviderFileUploadError,
                         ProviderGetUploadUrlError, )


PROVIDERS = ('backblaze', )
__all__ = ['Backblaze', 'ProviderAuthenticationError', 'PROVIDERS',
           'ProviderGetUploadUrlError', 'ProviderAuthorizationError',
           'ProviderFileUploadError']
